#ifndef SINGLEPP_UMBRELLA_HPP
#define SINGLEPP_UMBRELLA_HPP

/**
 * @file singlepp.hpp
 *
 * @brief Umbrella header for all includes.
 */

#include "Classifier.hpp"
#include "Markers.hpp"
#include "IntegratedScorer.hpp"
#include "IntegratedBuilder.hpp"
#include "ChooseClassicMarkers.hpp"
#include "load_references.hpp"

/**
 * @namespace singlepp
 * Cell type classification using the SingleR algorithm in C++.
 */

#endif
