# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

find_classic_markers <- function(nlabels, ngenes, labels, ref, de_n, nthreads) {
    .Call('_SingleR_find_classic_markers', PACKAGE = 'SingleR', nlabels, ngenes, labels, ref, de_n, nthreads)
}

integrate_build <- function(test_features, references, ref_ids, labels, prebuilt, nthreads) {
    .Call('_SingleR_integrate_build', PACKAGE = 'SingleR', test_features, references, ref_ids, labels, prebuilt, nthreads)
}

integrate_run <- function(test, results, integrated_build, quantile, nthreads) {
    .Call('_SingleR_integrate_run', PACKAGE = 'SingleR', test, results, integrated_build, quantile, nthreads)
}

#' @importFrom Rcpp sourceCpp
#' @useDynLib SingleR
prebuild <- function(ref, labels, markers, approximate, nthreads) {
    .Call('_SingleR_prebuild', PACKAGE = 'SingleR', ref, labels, markers, approximate, nthreads)
}

get_subset <- function(built) {
    .Call('_SingleR_get_subset', PACKAGE = 'SingleR', built)
}

is_valid_built <- function(built) {
    .Call('_SingleR_is_valid_built', PACKAGE = 'SingleR', built)
}

#' @importFrom Rcpp sourceCpp
#' @useDynLib SingleR
run <- function(test, subset, prebuilt, quantile, use_fine_tune, fine_tune_threshold, nthreads) {
    .Call('_SingleR_run', PACKAGE = 'SingleR', test, subset, prebuilt, quantile, use_fine_tune, fine_tune_threshold, nthreads)
}

